import os
import json
from enum import Enum
from typing import Any, Dict, Callable, Union
from types import ModuleType
import yaml
import jinja2
import requests
import importlib
from importlib.machinery import SourceFileLoader
from pathlib import Path
from pkg_resources import resource_string
from pydantic import BaseModel


class Serialization(Enum):
    yaml = "yaml"
    json = "json"
    orjson = "orjson"
    ujson = "ujson"
    jinja = "jinja"
    jinja2 = "jinja2"
    string = "string"
    raw = "raw"


class Protocol(Enum):
    file = "file"
    http = "http"
    https = "https"
    pkgdata = "pkgdata"
    env = "env"
    module = "module"
    s3 = "s3"
    python = "python"
    inline = "inline"


jinja_env = jinja2.Environment(enable_async=True, autoescape=True)


def passthrough(item: Any) -> Any:
    return item


def string(item: Any) -> Any:
    return str(item)


serializers: Dict[Serialization, Callable[[Any], Any]] = {
    Serialization.yaml: yaml.safe_load,
    Serialization.json: json.loads,
    Serialization.jinja: jinja_env.from_string,
    Serialization.jinja2: jinja_env.from_string,
    Serialization.string: string,
    Serialization.raw: passthrough,
}

try:
    import ujson

    serializers[Serialization.ujson] = ujson.loads
    jinja_env.policies["json.dumps_function"] = ujson.dumps  # type: ignore
except ImportError:
    # This lambda will raise an exception when the serializer is used; otherwise we should not crash
    serializers[Serialization.ujson] = lambda *a, **kw: raise_(
        ImportError("ujson must be installed to use in config_loaders")
    )

try:
    import orjson

    serializers[Serialization.orjson] = orjson.loads

    # orjson.dumps returns bytes, so we have to wrap & decode it
    def orjson_dumps(*args: Any, **kwargs: Any) -> Any:
        try:
            representation = orjson.dumps(*args, **kwargs)
        except TypeError:
            raise TypeError(f"Unable to dump objects using ORJSON: {args}, {kwargs}")
        try:
            return representation.decode()
        except Exception as e:
            raise e.__class__(
                f"Unable to decode ORJSON: {representation}. Original exception: {e}"  # type: ignore
            )

    jinja_env.policies["json.dumps_function"] = orjson_dumps  # type: ignore
    jinja_env.policies["json.dumps_kwargs"] = {"option": orjson.OPT_SORT_KEYS}  # type: ignore
except ImportError:
    # This lambda will raise an exception when the serializer is used; otherwise we should not crash
    serializers[Serialization.orjson] = lambda *a, **kw: raise_(
        ImportError("orjson must be installed to use in config_loaders")
    )

try:
    import boto3
except ImportError:
    boto3 = None


class Loadable(BaseModel):
    protocol: Protocol = Protocol.http
    serialization: Serialization = Serialization.yaml
    path: str

    def load(self, default: Any = None) -> Any:
        try:
            return loaders[self.protocol](self.path, self.serialization)
        except Exception:
            if default is not None:
                return default
            raise

    @staticmethod
    def from_legacy_fmt(s: str) -> "Loadable":
        if "://" not in s:
            return Loadable(
                protocol=Protocol.inline, serialization=Serialization.string, path=s
            )
        try:
            scheme, path = s.split("://")
        except ValueError:
            raise ValueError(s)
        try:
            p, s = scheme.split("+")
        except ValueError:
            p, s = scheme, "yaml"

        proto: Protocol = Protocol(p)
        serialization: Serialization = Serialization(s)
        if proto in (Protocol.python, Protocol.module):
            serialization = Serialization.raw
        if proto in (Protocol.http, Protocol.https):
            path = "://".join([proto.value, path])

        return Loadable(
            protocol=proto,
            serialization=serialization,
            path=path,
        )


def raise_(e: Exception) -> Exception:
    raise e


def load_file(path: str, loader: Serialization) -> Any:
    with open(path) as f:
        contents = f.read()
        try:
            return serializers[loader](contents)
        except FileNotFoundError:
            raise FileNotFoundError(f"Unable to load {path}")


def load_package_data(path: str, loader: Serialization) -> Any:
    pkg, pkg_file = path.split(":")
    data = resource_string(pkg, pkg_file)
    if isinstance(data, str):
        data = data.decode()
    return serializers[loader](data)


def load_http(path: str, loader: Serialization) -> Any:
    response = requests.get(path)
    response.raise_for_status()
    data = response.text
    return serializers[loader](data)


def load_env(variable: str, loader: Serialization = Serialization.raw) -> Any:
    data = os.getenv(variable)
    try:
        return serializers[loader](data)
    except AttributeError as e:
        raise AttributeError(
            f"Unable to read environment variable {variable}: {repr(e)}"
        )


def load_module(name: str, _: Serialization = Serialization.raw) -> Any:
    if ":" in name:
        mod, fn = name.rsplit(":", maxsplit=1)
    else:
        mod, fn = name, ""
    imported = importlib.import_module(mod)
    if fn != "":
        return getattr(imported, fn)
    return imported


def load_s3(path: str, loader: Serialization = Serialization.raw) -> Any:
    if isinstance(boto3, type(None)):
        raise ImportError(
            "boto3 must be installed to load S3 paths. Use ``pip install sovereign[boto]``"
        )
    bucket, key = path.split("/", maxsplit=1)
    s3 = boto3.client("s3")
    response = s3.get_object(Bucket=bucket, Key=key)
    data = "".join([chunk.decode() for chunk in response["Body"]])
    return serializers[loader](data)


def load_python(path: str, _: Serialization = Serialization.raw) -> ModuleType:
    p = str(Path(path).absolute())
    loader = SourceFileLoader(p, path=p)
    return loader.load_module(p)


def load_inline(path: str, _: Serialization = Serialization.raw) -> Any:
    return str(path)


loaders: Dict[Protocol, Callable[[str, Serialization], Union[str, Any]]] = {
    Protocol.file: load_file,
    Protocol.pkgdata: load_package_data,
    Protocol.http: load_http,
    Protocol.https: load_http,
    Protocol.env: load_env,
    Protocol.module: load_module,
    Protocol.s3: load_s3,
    Protocol.python: load_python,
    Protocol.inline: load_inline,
}



import json
import threading
from typing import Dict, Any, TypeVar, Optional, cast

import structlog
from structlog.exceptions import DropEvent
from structlog._config import BoundLoggerLazyProxy

from sovereign.schemas import SovereignConfigv2


LOG_QUEUE = threading.local()


def default_log_fmt() -> Dict[str, str]:
    return {
        "env": "{ENVIRONMENT}",
        "site": "{HOST}",
        "method": "{METHOD}",
        "uri_path": "{PATH}",
        "uri_query": "{QUERY}",
        "src_ip": "{SOURCE_IP}",
        "src_port": "{SOURCE_PORT}",
        "pid": "{PID}",
        "user_agent": "{USER_AGENT}",
        "bytes_in": "{BYTES_RX}",
        "bytes_out": "{BYTES_TX}",
        "status": "{STATUS_CODE}",
        "duration": "{DURATION}",
        "request_id": "{REQUEST_ID}",
        "resource_version": "{XDS_CLIENT_VERSION} -> {XDS_SERVER_VERSION}",
        "resource_names": "{XDS_RESOURCES}",
        "envoy_ver": "{XDS_ENVOY_VERSION}",
        "traceback": "{TRACEBACK}",
        "error": "{ERROR}",
        "detail": "{ERROR_DETAIL}",
    }


T = TypeVar("T", bound=Dict[str, Any])


_configured_log_fmt: Optional[Dict[str, str]] = None


class LoggerBootstrapper:
    def __init__(self, config: SovereignConfigv2) -> None:
        self.debug = config.debug
        self.app_logs_enabled = config.logging.application_logs.enabled
        self.access_logs_enabled = config.logging.access_logs.enabled
        self.ignore_empty = config.logging.access_logs.ignore_empty_fields
        self.log_fmt = config.logging.access_logs.log_fmt
        self.logger = self.bootstrap()

    def bootstrap(self) -> BoundLoggerLazyProxy:
        structlog.configure(
            processors=[
                self.access_logs_enabled_processor,
                self.debug_logs_processor,
                self.merge_in_threadlocal,
                self.format_log_fields,
                structlog.processors.JSONRenderer(),
            ]
        )
        logger = structlog.getLogger()
        self.configured_log_fmt = self.configured_log_format()
        return logger

    def access_logs_enabled_processor(
        self, logger: Any, method_name: str, event_dict: T
    ) -> T:
        if not self.access_logs_enabled:
            raise DropEvent
        return event_dict

    def debug_logs_processor(self, logger: Any, method_name: str, event_dict: T) -> T:
        if event_dict.get("level") == "debug" and not self.debug:
            raise DropEvent
        return event_dict

    def application_log(self, **kwargs: Any) -> None:
        if self.app_logs_enabled:
            self.logger.msg(**kwargs)

    def merge_in_threadlocal(self, logger: Any, method_name: str, event_dict: T) -> T:
        self._ensure_threadlocal()
        fields = LOG_QUEUE.fields.copy()
        fields.update(event_dict)
        return cast(T, fields)

    def clear_log_fields(self) -> None:
        LOG_QUEUE.fields = dict()

    def _ensure_threadlocal(self) -> None:
        if not hasattr(LOG_QUEUE, "fields"):
            LOG_QUEUE.fields = dict()

    def queue_log_fields(self, **kwargs: Any) -> None:
        self._ensure_threadlocal()
        LOG_QUEUE.fields.update(kwargs)

    def configured_log_format(
        self, format: Optional[Dict[str, str]] = _configured_log_fmt
    ) -> Dict[str, str]:
        if format is not None:
            return format
        if isinstance(self.log_fmt, str) and self.log_fmt != "":
            format = json.loads(self.log_fmt)
            if not isinstance(format, dict):
                raise RuntimeError(
                    f"Failed to parse log format as JSON: {self.log_fmt}"
                )
            return format
        return default_log_fmt()

    def format_log_fields(self, logger: Any, method_name: str, event_dict: T) -> T:
        formatted_dict: Dict[str, Any] = dict()
        for k, v in self.configured_log_format().items():
            try:
                value: str = v.format(**event_dict)
            except KeyError:
                value = "-"
            if value in (None, "-") and self.ignore_empty:
                continue
            formatted_dict[k] = value
        return cast(T, formatted_dict)



"""
Source abstract class
---------------------
This class can be subclassed, installed as an entry point, and then
used via configuration.

`todo entry point install guide`
"""
import abc
from typing import Any, Dict, List


class Source(abc.ABC):
    def __init__(self, config: Dict[str, Any], scope: str) -> None:
        """
        :param config: arbitrary configuration which can be used by the subclass
        """
        self.config = config
        self.scope = scope

    def setup(self) -> None:
        """
        Optional method which is invoked prior to the Source running self.get()
        """
        return None

    @abc.abstractmethod
    def get(self) -> List[Dict[str, Any]]:
        """
        Required method to retrieve data from an arbitrary source
        """
        raise NotImplementedError


class SourceImplementation(Source):
    def __call__(self, config: Dict[str, Any], scope: str) -> "SourceImplementation":
        return self

    def get(self) -> List[Dict[str, Any]]:
        return []



import yaml
from typing import Any, Dict, List
from collections import defaultdict
from fastapi import APIRouter, Query
from fastapi.responses import JSONResponse
from fastapi.encoders import jsonable_encoder
from sovereign import discovery, config, stats, poller, template_context
from sovereign.discovery import select_template
from sovereign.utils.mock import mock_discovery_request

router = APIRouter()


@router.get("/xds_dump", summary="Displays all xDS resources as JSON")
async def display_config(
    xds_type: str = Query(
        ..., title="xDS type", description="The type of request", example="clusters"
    ),
    service_cluster: str = Query(
        "*", title="The clients service cluster to emulate in this XDS request"
    ),
    resource_names: str = Query("", title="Envoy Resource names to request"),
    region: str = Query(
        None, title="The clients region to emulate in this XDS request"
    ),
    version: str = Query(
        "1.11.1", title="The clients envoy version to emulate in this XDS request"
    ),
) -> JSONResponse:
    ret: Dict[str, List[Dict[str, Any]]] = defaultdict(list)
    mock_request = mock_discovery_request(
        service_cluster=service_cluster,
        resource_names=resource_names.split(","),
        version=version,
        region=region,
    )
    response = await discovery.response(request=mock_request, xds_type=xds_type)
    ret["resources"] += response.resources
    safe_response = jsonable_encoder(ret)
    return JSONResponse(content=safe_response)


@router.get(
    "/debug_template",
    summary="Dumps raw representation of template output before serialization/extra processing",
)
async def debug_template(
    xds_type: str = Query(
        ..., title="xDS type", description="The type of request", example="clusters"
    ),
    service_cluster: str = Query(
        "*", title="The clients service cluster to emulate in this XDS request"
    ),
    resource_names: str = Query("", title="Envoy Resource names to request"),
    region: str = Query(
        None, title="The clients region to emulate in this XDS request"
    ),
    version: str = Query(
        "1.11.1", title="The clients envoy version to emulate in this XDS request"
    ),
) -> JSONResponse:
    mock_request = mock_discovery_request(
        service_cluster=service_cluster,
        resource_names=resource_names.split(","),
        version=version,
        region=region,
    )
    template = select_template(mock_request, xds_type)
    context = template_context.get_context(mock_request, template)
    context = dict(
        discovery_request=mock_request,
        host_header="debug",
        resource_names=mock_request.resources,
        **context,
    )
    raw_template_content = await template(**context)
    safe_response = jsonable_encoder(raw_template_content)
    return JSONResponse(content=safe_response)


@router.get(
    "/source_dump",
    summary="Displays all sources that this Sovereign has polled as JSON",
)
def instances(
    service_cluster: str = Query(
        "*", title="The clients service cluster to emulate in this XDS request"
    ),
    modified: str = Query(
        "yes",
        title="Whether the sources should run Modifiers/Global Modifiers prior to being returned",
    ),
) -> JSONResponse:
    node = mock_discovery_request(service_cluster=service_cluster).node
    args = {
        "modify": yaml.safe_load(modified),
        "node_value": poller.extract_node_key(node),
    }
    ret = poller.match_node(**args)
    safe_response = jsonable_encoder(ret)
    return JSONResponse(content=safe_response)


@router.get("/config", summary="Display the current Sovereign configuration")
def show_configuration() -> JSONResponse:
    safe_response = jsonable_encoder(config.show())
    return JSONResponse(content=safe_response)


@router.get("/stats", summary="Displays all metrics emitted and their counters")
def show_stats() -> JSONResponse:
    return JSONResponse(content=stats.emitted)



import random
from typing import Dict, Any, Optional, List
from copy import deepcopy
from starlette.exceptions import HTTPException
from sovereign import config
from sovereign.schemas import DiscoveryRequest
from sovereign.utils.templates import resolve

HARD_FAIL_ON_DNS_FAILURE = config.legacy_fields.dns_hard_fail
PRIORITY_MAPPING = config.legacy_fields.eds_priority_matrix
TOTAL_REGIONS = len(config.legacy_fields.regions or [])


def _upstream_kwargs(
    upstream: Dict[str, Any],
    proxy_region: Optional[str] = None,
    resolve_dns: bool = True,
    default_region: Optional[str] = None,
    hard_fail: Optional[bool] = HARD_FAIL_ON_DNS_FAILURE,
) -> Dict[str, Any]:
    try:
        ip_addresses = (
            resolve(upstream["address"]) if resolve_dns else [upstream["address"]]
        )
    except HTTPException:
        if hard_fail:
            raise
        ip_addresses = [upstream["address"]]
    return {
        "addrs": ip_addresses,
        "port": upstream["port"],
        "region": default_region or upstream.get("region", "unknown"),
        "zone": proxy_region,
    }


def total_zones(endpoints: List[Dict[str, Dict[str, Any]]]) -> int:
    """
    Returns the true unique number of zones, taking into account
    that multiple endpoints can have the same zone name.

    - us-west-1
    - us-west-1   == 2 zones
    - us-east-1

    - us-west-1
    - us-west-2   == 3 zones
    - us-east-2
    """
    zones = {e["locality"]["zone"] for e in endpoints}
    return len(zones)


def locality_lb_endpoints(
    upstreams: List[Dict[str, Any]],
    request: Optional[DiscoveryRequest] = None,
    resolve_dns: bool = True,
) -> List[Dict[str, Any]]:
    if request is None:
        proxy_region = None
    else:
        proxy_region = request.node.locality.zone

    kw_args = [_upstream_kwargs(u, proxy_region, resolve_dns) for u in upstreams]
    ret = [lb_endpoints(**kw) for kw in kw_args]

    if total_zones(ret) == 1:
        # Pointless to do zone-aware load-balancing for a single zone
        return ret

    upstreams_copy = deepcopy(upstreams)
    while total_zones(ret) < TOTAL_REGIONS:
        region = f"zone-padding-{total_zones(ret)}"
        try:
            upstream = upstreams_copy.pop()
        except IndexError:
            # When adding zone-padding, use a randomly selected upstream
            # However, the random selection should be consistent across control-planes
            # otherwise the version_info of the response will be constantly different
            random.seed(128)
            upstream = random.choice(upstreams)
        params = _upstream_kwargs(upstream, proxy_region, resolve_dns, region)
        ret.append(lb_endpoints(**params))
    return ret


def lb_endpoints(addrs: List[str], port: int, region: str, zone: str) -> Dict[str, Any]:
    """
    Creates an envoy endpoint.LbEndpoints proto

    :param addrs:  The IP addresses or hostname(s) of the upstream.
    :param port:   The port that the upstream should be accessed on.
    :param region: The region of the upstream.
    :param zone:   The region of the proxy asking for the endpoint configuration.
    """
    if PRIORITY_MAPPING is None:
        raise RuntimeError(
            "Tried to create LbEndpoints using the EDS utility,"
            " but no EDS priority matrix has been specified"
        )
    node_priorities = PRIORITY_MAPPING.get(zone, {})
    priority = node_priorities.get(region, 10)
    return {
        "priority": priority,
        "locality": {"zone": region},
        "lb_endpoints": [
            {
                "endpoint": {
                    "address": {
                        "socket_address": {
                            "address": addr,
                            "port_value": port,
                        }
                    }
                }
            }
            for addr in addrs
        ],
    }



from fastapi.exceptions import HTTPException
from cryptography.fernet import InvalidToken
from sovereign import config, stats, cipher_suite
from sovereign.schemas import DiscoveryRequest

AUTH_ENABLED = config.authentication.enabled


def validate_authentication_string(s: str) -> bool:
    try:
        password = cipher_suite.decrypt(s)
    except Exception:
        stats.increment("discovery.auth.failed")
        raise

    if password in config.passwords:
        stats.increment("discovery.auth.success")
        return True
    stats.increment("discovery.auth.failed")
    return False


def authenticate(request: DiscoveryRequest) -> None:
    if not AUTH_ENABLED:
        return
    if not cipher_suite.key_available:
        raise RuntimeError(
            "No Fernet key loaded, and auth is enabled. "
            "A fernet key must be provided via SOVEREIGN_ENCRYPTION_KEY. "
            "See https://vsyrakis.bitbucket.io/sovereign/docs/html/guides/encryption.html "
            "for more details"
        )
    try:
        encrypted_auth = request.node.metadata["auth"]
        with stats.timed("discovery.auth.ms"):
            assert validate_authentication_string(encrypted_auth)
    except KeyError:
        raise HTTPException(
            status_code=401,
            detail=f"Discovery request from {request.node.id} is missing auth field",
        )
    except (InvalidToken, AssertionError):
        raise HTTPException(
            status_code=401, detail="The authentication provided was invalid"
        )
    except Exception as e:
        description = getattr(e, "detail", "Unknown")
        raise HTTPException(
            status_code=400,
            detail=f"The authentication provided was malformed [Reason: {description}]",
        )



import zlib
from typing import Any


def compute_hash(*args: Any) -> str:
    """
    Creates a 'version hash' to be used in envoy Discovery Responses.
    """
    data: bytes = repr(args).encode()
    version_info = (
        zlib.adler32(data) & 0xFFFFFFFF
    )  # same numeric value across all py versions & platforms
    return str(version_info)



from typing import Dict, Any, Generator, Iterable, NoReturn, Optional
from copy import deepcopy
from fastapi import HTTPException
from sovereign.config_loader import Loadable
from sovereign.schemas import DiscoveryRequest, XdsTemplate
from sovereign.sources import SourcePoller
from sovereign.utils.crypto import CipherSuite
from sovereign.utils.version_info import compute_hash
from sovereign.utils.timer import poll_forever, poll_forever_cron


class TemplateContext:
    def __init__(
        self,
        refresh_rate: Optional[int],
        refresh_cron: Optional[str],
        configured_context: Dict[str, Loadable],
        poller: SourcePoller,
        encryption_suite: CipherSuite,
        disabled_suite: CipherSuite,
        logger: Any,
        stats: Any,
    ) -> None:
        self.poller = poller
        self.refresh_rate = refresh_rate
        self.refresh_cron = refresh_cron
        self.configured_context = configured_context
        self.crypto = encryption_suite
        self.disabled_suite = disabled_suite
        # initial load
        self.context = self.load_context_variables()
        self.checksum = compute_hash(self.context)
        self.logger = logger
        self.stats = stats

    async def start_refresh_context(self) -> NoReturn:
        if self.refresh_cron is not None:
            await poll_forever_cron(self.refresh_cron, self.refresh_context)
        elif self.refresh_rate is not None:
            await poll_forever(self.refresh_rate, self.refresh_context)

        raise RuntimeError("Failed to start refresh_context, this should never happen")

    async def refresh_context(self) -> None:
        try:
            self.context = self.load_context_variables()
            self.checksum = compute_hash(self.context)
            self.stats.increment("context.refresh.success")
        # pylint: disable=broad-except
        except Exception as e:
            self.logger(event=e)
            self.stats.increment("context.refresh.error")

    def load_context_variables(self) -> Dict[str, Any]:
        ret = dict()
        for k, v in self.configured_context.items():
            if isinstance(v, Loadable):
                ret[k] = v.load()
            elif isinstance(v, str):
                ret[k] = Loadable.from_legacy_fmt(v).load()
        if "crypto" not in ret:
            ret["crypto"] = self.crypto
        return ret

    def build_new_context_from_instances(self, node_value: str) -> Dict[str, Any]:
        matches = self.poller.match_node(node_value=node_value)
        ret = dict()
        for key, value in self.context.items():
            try:
                ret[key] = deepcopy(value)
            except TypeError:
                ret[key] = value

        to_add = dict()
        for scope, instances in matches.scopes.items():
            if scope in ("default", None):
                to_add["instances"] = instances
            else:
                to_add[scope] = instances
        if to_add == {}:
            raise HTTPException(
                detail=(
                    "This node does not match any instances! ",
                    "If node matching is enabled, check that the node "
                    "match key aligns with the source match key. "
                    "If you don't know what any of this is, disable "
                    "node matching via the config",
                ),
                status_code=400,
            )
        ret.update(to_add)
        return ret

    def safe(self, request: DiscoveryRequest) -> Dict[str, Any]:
        ret = self.build_new_context_from_instances(
            node_value=self.poller.extract_node_key(request.node),
        )
        # If the discovery request came from a mock, it will
        # typically contain this metadata key.
        # This means we should prevent any decryptable data
        # from ending up in the response.
        if request.hide_private_keys:
            ret["crypto"] = self.disabled_suite
        return ret

    def get_context(
        self, request: DiscoveryRequest, template: XdsTemplate
    ) -> Dict[str, Any]:
        ret = self.safe(request)
        if not template.is_python_source:
            keys_to_remove = self.unused_variables(list(ret), template.jinja_variables)
            for key in keys_to_remove:
                ret.pop(key, None)
        return ret

    @staticmethod
    def unused_variables(
        keys: Iterable[str], variables: Iterable[str]
    ) -> Generator[str, None, None]:
        for key in keys:
            if key not in variables:
                yield key

    def get(self, *args: Any, **kwargs: Any) -> Any:
        return self.context.get(*args, **kwargs)



import zlib
import warnings
import multiprocessing
from collections import defaultdict
from enum import Enum
from pydantic import (
    BaseModel,
    Field,
    BaseSettings,
    SecretStr,
    validator,
    root_validator,
)
from typing import List, Any, Dict, Union, Optional, Tuple, Type
from types import ModuleType
from jinja2 import meta, Template
from fastapi.responses import JSONResponse
from sovereign.config_loader import jinja_env, Serialization, Protocol, Loadable
from sovereign.utils.version_info import compute_hash
from croniter import croniter, CroniterBadCronError

missing_arguments = {"missing", "positional", "arguments:"}

JsonResponseClass: Type[JSONResponse] = JSONResponse
# pylint: disable=unused-import
try:
    import orjson
    from fastapi.responses import ORJSONResponse

    JsonResponseClass = ORJSONResponse
except ImportError:
    try:
        import ujson
        from fastapi.responses import UJSONResponse

        JsonResponseClass = UJSONResponse
    except ImportError:
        pass


class CacheStrategy(str, Enum):
    context = "context"
    content = "content"


class SourceData(BaseModel):
    scopes: Dict[str, List[Dict[str, Any]]] = defaultdict(list)


class ConfiguredSource(BaseModel):
    type: str
    config: Dict[str, Any]
    scope: str = "default"  # backward compatibility


class StatsdConfig(BaseModel):
    host: str = "127.0.0.1"
    port: int = 8125
    tags: Dict[str, Union[Loadable, str]] = dict()
    namespace: str = "sovereign"
    enabled: bool = False
    use_ms: bool = True

    @validator("tags", pre=True)
    def load_tags(cls, v: Dict[str, Union[Loadable, str]]) -> Dict[str, Any]:
        ret = dict()
        for key, value in v.items():
            if isinstance(value, dict):
                ret[key] = Loadable(**value).load()
            elif isinstance(value, str):
                ret[key] = Loadable.from_legacy_fmt(value).load()
            else:
                raise ValueError(f"Received an invalid tag for statsd: {value}")
        return ret


class XdsTemplate:
    def __init__(self, path: Union[str, Loadable]) -> None:
        if isinstance(path, str):
            self.loadable: Loadable = Loadable.from_legacy_fmt(path)
        elif isinstance(path, Loadable):
            self.loadable = path
        self.is_python_source = self.loadable.protocol == Protocol.python
        self.source = self.load_source()
        self.checksum = zlib.adler32(self.source.encode())
        template_ast = jinja_env.parse(self.source)
        self.jinja_variables = meta.find_undeclared_variables(template_ast)  # type: ignore

    async def __call__(self, *args: Any, **kwargs: Any) -> Union[Dict[str, Any], str]:
        if not hasattr(self, "code"):
            self.code: Union[Template, ModuleType] = self.loadable.load()
        if isinstance(self.code, ModuleType):
            try:
                return {"resources": list(self.code.call(*args, **kwargs))}  # type: ignore
            except TypeError as e:
                if not set(str(e).split()).issuperset(missing_arguments):
                    raise e
                message_start = str(e).find(":")
                missing_args = str(e)[message_start + 2 :]
                supplied_args = list(kwargs.keys())
                raise TypeError(
                    f"Tried to render a template using partial arguments. "
                    f"Missing args: {missing_args}. Supplied args: {args} "
                    f"Supplied keyword args: {supplied_args}"
                )
        else:
            return await self.code.render_async(*args, **kwargs)

    def load_source(self) -> str:
        if self.loadable.serialization in (Serialization.jinja, Serialization.jinja2):
            # The Jinja2 template serializer does not properly set a name
            # for the loaded template.
            # The repr for the template prints out as the memory address
            # This makes it really hard to generate a consistent version_info string
            # in rendered configuration.
            # For this reason, we re-load the template as a string instead, and create a checksum.
            old_serialization = self.loadable.serialization
            self.loadable.serialization = Serialization("string")
            ret = self.loadable.load()
            self.loadable.serialization = old_serialization
            return str(ret)
        elif self.is_python_source:
            # If the template specified is a python source file,
            # we can simply read and return the source of it.
            old_protocol = self.loadable.protocol
            old_serialization = self.loadable.serialization
            self.loadable.protocol = Protocol("inline")
            self.loadable.serialization = Serialization("string")
            ret = self.loadable.load()
            self.loadable.protocol = old_protocol
            self.loadable.serialization = old_serialization
            return str(ret)
        else:
            # The only other supported serializers are string, yaml, and json
            # So it should be safe to create this checksum off
            return str(self.source)


class ProcessedTemplate:
    def __init__(
        self,
        resources: List[Dict[str, Any]],
        type_url: str,
        version_info: Optional[str],
    ) -> None:
        for resource in resources:
            if "@type" not in resource:
                resource["@type"] = type_url
        self.resources = resources
        self.version_info = version_info

    @property
    def version(self) -> str:
        return self.version_info or compute_hash(self.resources)

    @property
    def rendered(self) -> bytes:
        return JsonResponseClass().render(
            content={
                "version_info": self.version,
                "resources": self.resources,
            }
        )

    def deserialize_resources(self) -> List[Dict[str, Any]]:
        return self.resources


class Locality(BaseModel):
    region: str = Field(None)
    zone: str = Field(None)
    sub_zone: str = Field(None)


class SemanticVersion(BaseModel):
    major_number: int = 0
    minor_number: int = 0
    patch: int = 0

    def __str__(self) -> str:
        return f"{self.major_number}.{self.minor_number}.{self.patch}"


class BuildVersion(BaseModel):
    version: SemanticVersion = SemanticVersion()
    metadata: Dict[str, Any] = {}


class Extension(BaseModel):
    name: Optional[str] = None
    category: Optional[str] = None
    version: Optional[BuildVersion] = None
    disabled: Optional[bool] = None


class Node(BaseModel):
    id: str = Field("-", title="Hostname")
    cluster: str = Field(
        ...,
        title="Envoy service-cluster",
        description="The ``--service-cluster`` configured by the Envoy client",
    )
    metadata: Dict[str, Any] = Field(default_factory=dict, title="Key:value metadata")
    locality: Locality = Field(Locality(), title="Locality")
    build_version: str = Field(
        None,  # Optional in the v3 Envoy API
        title="Envoy build/release version string",
        description="Used to identify what version of Envoy the "
        "client is running, and what config to provide in response",
    )
    user_agent_name: str = "envoy"
    user_agent_version: str = ""
    user_agent_build_version: BuildVersion = BuildVersion()
    extensions: List[Extension] = []
    client_features: List[str] = []

    @property
    def common(self) -> Tuple[str, str, str, BuildVersion, Locality]:
        """
        Returns fields that are the same in adjacent proxies
        ie. proxies that are part of the same logical group
        """
        return (
            self.cluster,
            self.build_version,
            self.user_agent_version,
            self.user_agent_build_version,
            self.locality,
        )


class Resources(List[str]):
    """
    Acts like a regular list except it returns True
    for all membership tests when empty.
    """

    def __contains__(self, item: object) -> bool:
        if len(self) == 0:
            return True
        return item in list(self)


class DiscoveryRequest(BaseModel):
    node: Node = Field(..., title="Node information about the envoy proxy")
    version_info: str = Field(
        "0", title="The version of the envoy clients current configuration"
    )
    resource_names: Resources = Field(
        Resources(), title="List of requested resource names"
    )
    hide_private_keys: bool = False
    type_url: str = Field(
        None, title="The corresponding type_url for the requested resource"
    )
    desired_controlplane: str = Field(
        None, title="The host header provided in the Discovery Request"
    )

    @property
    def envoy_version(self) -> str:
        try:
            version = str(self.node.user_agent_build_version.version)
            assert version != "0.0.0"
        except AssertionError:
            try:
                build_version = self.node.build_version
                _, version, *_ = build_version.split("/")
            except (AttributeError, ValueError):
                # TODO: log/metric this?
                return "default"
        return version

    @property
    def resources(self) -> Resources:
        return Resources(self.resource_names)

    @property
    def uid(self) -> str:
        return compute_hash(
            self.resources,
            self.node.common,
            self.desired_controlplane,
        )


class DiscoveryResponse(BaseModel):
    version_info: str = Field(
        ..., title="The version of the configuration in the response"
    )
    resources: List[Any] = Field(..., title="The requested configuration resources")


class SovereignAsgiConfig(BaseSettings):
    host: str = "0.0.0.0"
    port: int = 8080
    keepalive: int = 5
    workers: int = multiprocessing.cpu_count() * 2 + 1
    threads: int = 1
    reuse_port: bool = True
    preload_app: bool = False
    log_level: str = "warning"
    worker_class: str = "uvicorn.workers.UvicornWorker"
    worker_timeout: int = 30

    class Config:
        fields = {
            "host": {"env": "SOVEREIGN_HOST"},
            "port": {"env": "SOVEREIGN_PORT"},
            "keepalive": {"env": "SOVEREIGN_KEEPALIVE"},
            "workers": {"env": "SOVEREIGN_WORKERS"},
            "threads": {"env": "SOVEREIGN_THREADS"},
            "preload_app": {"env": "SOVEREIGN_PRELOAD"},
            "worker_timeout": {"env": "SOVEREIGN_WORKER_TIMEOUT"},
        }

    def as_gunicorn_conf(self) -> Dict[str, Any]:
        return {
            "bind": ":".join(map(str, [self.host, self.port])),
            "keepalive": self.keepalive,
            "reuse_port": self.reuse_port,
            "preload_app": self.preload_app,
            "loglevel": self.log_level,
            "timeout": self.worker_timeout,
            "threads": self.threads,
            "workers": self.workers,
            "worker_class": self.worker_class,
        }


class SovereignConfig(BaseSettings):
    sources: List[ConfiguredSource]
    templates: Dict[str, Dict[str, Union[str, Loadable]]]
    template_context: Dict[str, Any] = {}
    eds_priority_matrix: Dict[str, Dict[str, str]] = {}
    modifiers: List[str] = []
    global_modifiers: List[str] = []
    regions: List[str] = []
    statsd: StatsdConfig = StatsdConfig()
    auth_enabled: bool = False
    auth_passwords: str = ""
    encryption_key: str = ""
    environment: str = "local"
    debug_enabled: bool = False
    sentry_dsn: str = ""
    node_match_key: str = "cluster"
    node_matching: bool = True
    source_match_key: str = "service_clusters"
    sources_refresh_rate: int = 30
    cache_strategy: str = "context"
    refresh_context: bool = False
    context_refresh_rate: Optional[int]
    context_refresh_cron: Optional[str]
    dns_hard_fail: bool = False
    enable_application_logs: bool = False
    enable_access_logs: bool = True
    log_fmt: Optional[str] = ""
    ignore_empty_log_fields: bool = False

    class Config:
        fields = {
            "auth_enabled": {"env": "SOVEREIGN_AUTH_ENABLED"},
            "auth_passwords": {"env": "SOVEREIGN_AUTH_PASSWORDS"},
            "encryption_key": {"env": "SOVEREIGN_ENCRYPTION_KEY"},
            "environment": {"env": "SOVEREIGN_ENVIRONMENT"},
            "debug_enabled": {"env": "SOVEREIGN_DEBUG_ENABLED"},
            "sentry_dsn": {"env": "SOVEREIGN_SENTRY_DSN"},
            "node_match_key": {"env": "SOVEREIGN_NODE_MATCH_KEY"},
            "node_matching": {"env": "SOVEREIGN_NODE_MATCHING"},
            "source_match_key": {"env": "SOVEREIGN_SOURCE_MATCH_KEY"},
            "sources_refresh_rate": {"env": "SOVEREIGN_SOURCES_REFRESH_RATE"},
            "cache_strategy": {"env": "SOVEREIGN_CACHE_STRATEGY"},
            "refresh_context": {"env": "SOVEREIGN_REFRESH_CONTEXT"},
            "context_refresh_rate": {"env": "SOVEREIGN_CONTEXT_REFRESH_RATE"},
            "context_refresh_cron": {"env": "SOVEREIGN_CONTEXT_REFRESH_CRON"},
            "dns_hard_fail": {"env": "SOVEREIGN_DNS_HARD_FAIL"},
            "enable_application_logs": {"env": "SOVEREIGN_ENABLE_APPLICATION_LOGS"},
            "enable_access_logs": {"env": "SOVEREIGN_ENABLE_ACCESS_LOGS"},
            "log_fmt": {"env": "SOVEREIGN_LOG_FORMAT"},
            "ignore_empty_fields": {"env": "SOVEREIGN_LOG_IGNORE_EMPTY"},
        }

    @property
    def passwords(self) -> List[str]:
        return self.auth_passwords.split(",") or []

    def xds_templates(self) -> Dict[str, Dict[str, XdsTemplate]]:
        ret: Dict[str, Dict[str, XdsTemplate]] = {
            "__any__": {}
        }  # Special key to hold templates from all versions
        for version, templates in self.templates.items():
            loaded_templates = {
                _type: XdsTemplate(path=path) for _type, path in templates.items()
            }
            ret[str(version)] = loaded_templates
            ret["__any__"].update(loaded_templates)
        return ret

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        kwargs = [f"{k}={v}" for k, v in self.show().items()]
        return f"SovereignConfig({kwargs})"

    def show(self) -> Dict[str, Any]:
        safe_items = dict()
        for key, value in self.__dict__.items():
            if key in ["auth_passwords", "encryption_key", "passwords", "sentry_dsn"]:
                value = "redacted"
            safe_items[key] = value
        return safe_items


class TemplateSpecification(BaseModel):
    type: str
    spec: Loadable


class NodeMatching(BaseSettings):
    enabled: bool = True
    source_key: str = "service_clusters"
    node_key: str = "cluster"

    class Config:
        fields = {
            "enabled": {"env": "SOVEREIGN_NODE_MATCHING_ENABLED"},
            "source_key": {"env": "SOVEREIGN_SOURCE_MATCH_KEY"},
            "node_key": {"env": "SOVEREIGN_NODE_MATCH_KEY"},
        }


class AuthConfiguration(BaseSettings):
    enabled: bool = False
    auth_passwords: SecretStr = SecretStr("")
    encryption_key: SecretStr = SecretStr("")

    class Config:
        fields = {
            "enabled": {"env": "SOVEREIGN_AUTH_ENABLED"},
            "auth_passwords": {"env": "SOVEREIGN_AUTH_PASSWORDS"},
            "encryption_key": {"env": "SOVEREIGN_ENCRYPTION_KEY"},
        }


class ApplicationLogConfiguration(BaseSettings):
    enabled: bool = False
    # currently only support /dev/stdout as JSON

    class Config:
        fields = {
            "enabled": {"env": "SOVEREIGN_ENABLE_APPLICATION_LOGS"},
        }


class AccessLogConfiguration(BaseSettings):
    enabled: bool = True
    log_fmt: Optional[str] = None
    ignore_empty_fields: bool = False

    class Config:
        fields = {
            "enabled": {"env": "SOVEREIGN_ENABLE_ACCESS_LOGS"},
            "log_fmt": {"env": "SOVEREIGN_LOG_FORMAT"},
            "ignore_empty_fields": {"env": "SOVEREIGN_LOG_IGNORE_EMPTY"},
        }


class LoggingConfiguration(BaseSettings):
    application_logs: ApplicationLogConfiguration = ApplicationLogConfiguration()
    access_logs: AccessLogConfiguration = AccessLogConfiguration()


class ContextConfiguration(BaseSettings):
    context: Dict[str, Loadable] = {}
    refresh: bool = False
    refresh_rate: Optional[int] = None
    refresh_cron: Optional[str] = None

    @staticmethod
    def context_from_legacy(context: Dict[str, str]) -> Dict[str, Loadable]:
        ret = dict()
        for key, value in context.items():
            ret[key] = Loadable.from_legacy_fmt(value)
        return ret

    @root_validator(pre=False)
    def validate_single_use_refresh_method(
        cls, values: Dict[str, Any]
    ) -> Dict[str, Any]:
        refresh_rate = values.get("refresh_rate")
        refresh_cron = values.get("refresh_cron")

        if (refresh_rate is not None) and (refresh_cron is not None):
            raise RuntimeError(
                f"Only one of SOVEREIGN_CONTEXT_REFRESH_RATE or SOVEREIGN_CONTEXT_REFRESH_CRON can be defined. Got {refresh_rate=} and {refresh_cron=}"
            )
        return values

    @root_validator
    def set_default_refresh_rate(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        refresh_rate = values.get("refresh_rate")
        refresh_cron = values.get("refresh_cron")

        if (refresh_rate is None) and (refresh_cron is None):
            values["refresh_rate"] = 3600
        return values

    @validator("refresh_cron")
    def validate_refresh_cron(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        if not croniter.is_valid(v):
            raise CroniterBadCronError(f"'{v}' is not a valid cron expression")
        return v

    class Config:
        fields = {
            "refresh": {"env": "SOVEREIGN_REFRESH_CONTEXT"},
            "refresh_rate": {"env": "SOVEREIGN_CONTEXT_REFRESH_RATE"},
            "refresh_cron": {"env": "SOVEREIGN_CONTEXT_REFRESH_CRON"},
        }


class SourcesConfiguration(BaseSettings):
    refresh_rate: int = 30
    cache_strategy: CacheStrategy = CacheStrategy.context

    class Config:
        fields = {
            "refresh_rate": {"env": "SOVEREIGN_SOURCES_REFRESH_RATE"},
            "cache_strategy": {"env": "SOVEREIGN_CACHE_STRATEGY"},
        }


class LegacyConfig(BaseSettings):
    regions: Optional[List[str]] = None
    eds_priority_matrix: Optional[Dict[str, Dict[str, str]]] = None
    dns_hard_fail: Optional[bool] = None
    environment: Optional[str] = None

    @validator("regions")
    def regions_is_set(cls, v: Optional[List[str]]) -> List[str]:
        if v is not None:
            warnings.warn(
                "Setting regions via config is deprecated. "
                "It is suggested to use a modifier or template "
                "logic in order to achieve the same goal.",
                DeprecationWarning,
            )
            return v
        else:
            return []

    @validator("eds_priority_matrix")
    def eds_priority_matrix_is_set(
        cls, v: Optional[Dict[str, Dict[str, Any]]]
    ) -> Dict[str, Dict[str, Any]]:
        if v is not None:
            warnings.warn(
                "Setting eds_priority_matrix via config is deprecated. "
                "It is suggested to use a modifier or template "
                "logic in order to achieve the same goal.",
                DeprecationWarning,
            )
            return v
        else:
            return {}

    @validator("dns_hard_fail")
    def dns_hard_fail_is_set(cls, v: Optional[bool]) -> bool:
        if v is not None:
            warnings.warn(
                "Setting dns_hard_fail via config is deprecated. "
                "It is suggested to supply a module that can perform "
                "dns resolution to template_context, so that it can "
                "be used via templates instead.",
                DeprecationWarning,
            )
            return v
        else:
            return False

    @validator("environment")
    def environment_is_set(cls, v: Optional[str]) -> Optional[str]:
        if v is not None:
            warnings.warn(
                "Setting environment via config is deprecated. "
                "It is suggested to configure this value through log_fmt "
                "instead.",
                DeprecationWarning,
            )
            return v
        else:
            return None

    class Config:
        fields = {
            "dns_hard_fail": {"env": "SOVEREIGN_DNS_HARD_FAIL"},
            "environment": {"env": "SOVEREIGN_ENVIRONMENT"},
        }


class SovereignConfigv2(BaseSettings):
    sources: List[ConfiguredSource]
    templates: Dict[str, List[TemplateSpecification]]
    source_config: SourcesConfiguration = SourcesConfiguration()
    modifiers: List[str] = []
    global_modifiers: List[str] = []
    template_context: ContextConfiguration = ContextConfiguration()
    matching: NodeMatching = NodeMatching()
    authentication: AuthConfiguration = AuthConfiguration()
    logging: LoggingConfiguration = LoggingConfiguration()
    statsd: StatsdConfig = StatsdConfig()
    sentry_dsn: SecretStr = SecretStr("")
    debug: bool = False
    legacy_fields: LegacyConfig = LegacyConfig()

    class Config:
        fields = {
            "sentry_dsn": {"env": "SOVEREIGN_SENTRY_DSN"},
            "debug": {"env": "SOVEREIGN_DEBUG"},
        }

    @property
    def passwords(self) -> List[str]:
        return self.authentication.auth_passwords.get_secret_value().split(",") or []

    def xds_templates(self) -> Dict[str, Dict[str, XdsTemplate]]:
        ret: Dict[str, Dict[str, XdsTemplate]] = {
            "__any__": {}
        }  # Special key to hold templates from all versions
        for version, template_specs in self.templates.items():
            loaded_templates = {
                template.type: XdsTemplate(path=template.spec)
                for template in template_specs
            }
            ret[str(version)] = loaded_templates
            ret["__any__"].update(loaded_templates)
        return ret

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f"SovereignConfigv2({self.dict()})"

    def show(self) -> Dict[str, Any]:
        return self.dict()

    @staticmethod
    def from_legacy_config(other: SovereignConfig) -> "SovereignConfigv2":
        new_templates = dict()
        for version, templates in other.templates.items():
            specs = list()
            for type, path in templates.items():
                if isinstance(path, str):
                    specs.append(
                        TemplateSpecification(
                            type=type, spec=Loadable.from_legacy_fmt(path)
                        )
                    )
                else:
                    # Just in case? Although this shouldn't happen
                    specs.append(TemplateSpecification(type=type, spec=path))
            new_templates[str(version)] = specs

        return SovereignConfigv2(
            sources=other.sources,
            templates=new_templates,
            source_config=SourcesConfiguration(
                refresh_rate=other.sources_refresh_rate,
                cache_strategy=other.cache_strategy,
            ),
            modifiers=other.modifiers,
            global_modifiers=other.global_modifiers,
            template_context=ContextConfiguration(
                context=ContextConfiguration.context_from_legacy(
                    other.template_context
                ),
                refresh=other.refresh_context,
                refresh_rate=other.context_refresh_rate,
                refresh_cron=other.context_refresh_cron,
            ),
            matching=NodeMatching(
                enabled=other.node_matching,
                source_key=other.source_match_key,
                node_key=other.node_match_key,
            ),
            authentication=AuthConfiguration(
                enabled=other.auth_enabled,
                auth_passwords=SecretStr(other.auth_passwords),
                encryption_key=SecretStr(other.encryption_key),
            ),
            logging=LoggingConfiguration(
                application_logs=ApplicationLogConfiguration(
                    enabled=other.enable_application_logs,
                ),
                access_logs=AccessLogConfiguration(
                    enabled=other.enable_access_logs,
                    log_fmt=other.log_fmt,
                    ignore_empty_fields=other.ignore_empty_log_fields,
                ),
            ),
            statsd=other.statsd,
            sentry_dsn=SecretStr(other.sentry_dsn),
            debug=other.debug_enabled,
            legacy_fields=LegacyConfig(
                regions=other.regions,
                eds_priority_matrix=other.eds_priority_matrix,
                dns_hard_fail=other.dns_hard_fail,
                environment=other.environment,
            ),
        )



import os
from contextvars import ContextVar
from typing import Type, Any, Mapping
from pkg_resources import get_distribution, resource_filename

from fastapi.responses import JSONResponse
from starlette.templating import Jinja2Templates
from pydantic.error_wrappers import ValidationError

from sovereign.schemas import (
    SovereignAsgiConfig,
    SovereignConfig,
    SovereignConfigv2,
)
from sovereign import config_loader
from sovereign.logs import LoggerBootstrapper
from sovereign.statistics import configure_statsd
from sovereign.utils.dictupdate import merge  # type: ignore
from sovereign.sources import SourcePoller
from sovereign.context import TemplateContext
from sovereign.utils.crypto import create_cipher_suite


json_response_class: Type[JSONResponse] = JSONResponse
try:
    import orjson
    from fastapi.responses import ORJSONResponse

    json_response_class = ORJSONResponse
except ImportError:
    try:
        import ujson
        from fastapi.responses import UJSONResponse

        json_response_class = UJSONResponse
    except ImportError:
        pass


def parse_raw_configuration(path: str) -> Mapping[Any, Any]:
    ret: Mapping[Any, Any] = dict()
    for p in path.split(","):
        spec = config_loader.Loadable.from_legacy_fmt(p)
        ret = merge(obj_a=ret, obj_b=spec.load(), merge_lists=True)
    return ret


_request_id_ctx_var: ContextVar[str] = ContextVar("request_id", default="")


def get_request_id() -> str:
    return _request_id_ctx_var.get()


__version__ = get_distribution("sovereign").version
config_path = os.getenv("SOVEREIGN_CONFIG", "file:///etc/sovereign.yaml")
html_templates = Jinja2Templates(resource_filename("sovereign", "templates"))

try:
    config = SovereignConfigv2(**parse_raw_configuration(config_path))
except ValidationError:
    old_config = SovereignConfig(**parse_raw_configuration(config_path))
    config = SovereignConfigv2.from_legacy_config(old_config)
asgi_config = SovereignAsgiConfig()
XDS_TEMPLATES = config.xds_templates()

logs = LoggerBootstrapper(config)
stats = configure_statsd(config=config.statsd)
poller = SourcePoller(
    sources=config.sources,
    matching_enabled=config.matching.enabled,
    node_match_key=config.matching.node_key,
    source_match_key=config.matching.source_key,
    source_refresh_rate=config.source_config.refresh_rate,
    logger=logs.application_log,
    stats=stats,
)

encryption_key = config.authentication.encryption_key.get_secret_value().encode()
cipher_suite = create_cipher_suite(key=encryption_key, logger=logs)

template_context = TemplateContext(
    refresh_rate=config.template_context.refresh_rate,
    refresh_cron=config.template_context.refresh_cron,
    configured_context=config.template_context.context,
    poller=poller,
    encryption_suite=cipher_suite,
    disabled_suite=create_cipher_suite(b"", logs),
    logger=logs.application_log,
    stats=stats,
)
poller.lazy_load_modifiers(config.modifiers)
poller.lazy_load_global_modifiers(config.global_modifiers)



from typing import Optional, Dict, List
from random import randint
from sovereign.schemas import DiscoveryRequest, Node, Locality


def mock_discovery_request(
    service_cluster: Optional[str] = None,
    resource_names: Optional[List[str]] = None,
    region: str = "none",
    version: str = "1.11.1",
    metadata: Optional[Dict[str, str]] = None,
) -> DiscoveryRequest:
    request = DiscoveryRequest(
        node=Node(
            id="mock",
            cluster=service_cluster or "",
            build_version=f"e5f864a82d4f27110359daa2fbdcb12d99e415b9/{version}/Clean/RELEASE",
            locality=Locality(zone=region),
        ),
        version_info=str(randint(100000, 1000000000)),
        resource_names=resource_names or [],
        hide_private_keys=True,
    )
    if isinstance(metadata, dict):
        request.node.metadata = metadata
    return request



from typing import Dict
from pydantic import BaseModel, Field
from fastapi import APIRouter, Body
from fastapi.responses import JSONResponse
from sovereign import json_response_class, cipher_suite
from sovereign.utils.crypto import generate_key

router = APIRouter()


class EncryptionRequest(BaseModel):
    data: str = Field(..., title="Text to be encrypted", min_length=1, max_length=65535)
    key: str = Field(
        None,
        title="Optional Fernet encryption key to use to encrypt",
        min_length=44,
        max_length=44,
    )


class DecryptionRequest(BaseModel):
    data: str = Field(..., title="Text to be decrypted", min_length=1, max_length=65535)
    key: str = Field(
        ...,
        title="Fernet encryption key to use to decrypt",
        min_length=44,
        max_length=44,
    )


class DecryptableRequest(BaseModel):
    data: str = Field(..., title="Text to be decrypted", min_length=1, max_length=65535)


@router.post(
    "/decrypt",
    summary="Decrypt provided encrypted data using a provided key",
    response_class=json_response_class,
)
async def _decrypt(request: DecryptionRequest = Body(None)) -> Dict[str, str]:
    return {"result": cipher_suite.decrypt(request.data, request.key)}


@router.post(
    "/encrypt",
    summary="Encrypt provided data using this servers key",
    response_class=json_response_class,
)
async def _encrypt(request: EncryptionRequest = Body(None)) -> Dict[str, str]:
    return {"result": cipher_suite.encrypt(data=request.data, key=request.key)}


@router.post(
    "/decryptable",
    summary="Check whether data is decryptable by this server",
    response_class=json_response_class,
)
async def _decryptable(request: DecryptableRequest = Body(None)) -> JSONResponse:
    cipher_suite.decrypt(request.data)
    return json_response_class({})


@router.get(
    "/generate_key",
    summary="Generate a new asymmetric encryption key",
    response_class=json_response_class,
)
def _generate_key() -> Dict[str, str]:
    return {"result": generate_key()}



import asyncio
import datetime as dt
from datetime import datetime
from typing import Any, Callable, Coroutine, NoReturn

from croniter import croniter


def wait_until(dt: datetime) -> float:
    now = datetime.now()
    sleep_time = (dt - now).total_seconds()
    return sleep_time


async def poll_forever(
    delay: int, coro: Callable[[], Coroutine[Any, Any, Any]]
) -> NoReturn:
    while True:
        await coro()
        await asyncio.sleep(delay)


async def poll_forever_cron(
    cron_expression: str, coro: Callable[[], Coroutine[Any, Any, Any]]
) -> NoReturn:
    croniter_iter = croniter(cron_expression)
    while True:
        await coro()
        next_datetime = croniter_iter.get_next(dt.datetime)
        delay = wait_until(next_datetime)
        await asyncio.sleep(delay)



import asyncio
import traceback
from copy import deepcopy
from datetime import timedelta, datetime
from pkg_resources import iter_entry_points, EntryPoint
from typing import Iterable, Any, Dict, List, Union, Type, Optional

from glom import glom, PathAccessError

from sovereign.sources.lib import Source
from sovereign.modifiers.lib import Modifier, GlobalModifier
from sovereign.schemas import (
    ConfiguredSource,
    SourceData,
    Node,
)


def is_debug_request(v: str, debug: bool = False) -> bool:
    return v == "" and debug


def is_wildcard(v: List[str]) -> bool:
    return v in [["*"], "*", ("*",)]


def contains(container: Iterable[Any], item: Any) -> bool:
    return item in container


Mods = Dict[str, Type[Modifier]]
GMods = Dict[str, Type[GlobalModifier]]


class SourcePoller:
    def __init__(
        self,
        sources: List[ConfiguredSource],
        matching_enabled: bool,
        node_match_key: str,
        source_match_key: str,
        source_refresh_rate: int,
        logger: Any,
        stats: Any,
    ):
        self.matching_enabled = matching_enabled
        self.node_match_key = node_match_key
        self.source_match_key = source_match_key
        self.source_refresh_rate = source_refresh_rate
        self.logger = logger
        self.stats = stats

        self.entry_points: Dict[str, List[EntryPoint]] = dict()
        for entry_point in ("sources", "modifiers", "global_modifiers"):
            self.entry_points[entry_point] = list(
                iter_entry_points(f"sovereign.{entry_point}")
            )

        self.source_classes: Dict[str, Type[Source]] = {
            e.name: e.load() for e in self.entry_points["sources"]
        }
        self.sources = [self.setup_source(s) for s in sources]
        if not self.sources:
            raise RuntimeError("No data sources available!")

        # These have to be loaded later to avoid circular imports
        self.modifiers: Mods = dict()
        self.global_modifiers: GMods = dict()

        # initially set data and modify
        self.source_data = self.refresh()
        self.last_updated = datetime.now()
        self.instance_count = 0

    @property
    def data_is_stale(self) -> bool:
        return self.last_updated < datetime.now() - timedelta(minutes=2)

    def setup_source(self, configured_source: ConfiguredSource) -> Source:
        source_class = self.source_classes[configured_source.type]
        source = source_class(
            config=configured_source.config,
            scope=configured_source.scope,
        )
        source.setup()
        return source

    def lazy_load_modifiers(self, modifiers: List[str]) -> None:
        if len(self.modifiers) == len(modifiers):
            return
        self.modifiers = self.load_modifier_entrypoints(
            self.entry_points["modifiers"], modifiers
        )

    def lazy_load_global_modifiers(self, global_modifiers: List[str]) -> None:
        if len(self.global_modifiers) == len(global_modifiers):
            return
        self.global_modifiers = self.load_global_modifier_entrypoints(
            self.entry_points["global_modifiers"], global_modifiers
        )

    def load_modifier_entrypoints(
        self, entry_points: Iterable[EntryPoint], configured_modifiers: List[str]
    ) -> Dict[str, Type[Modifier]]:
        ret = dict()
        for entry_point in entry_points:
            if entry_point.name in configured_modifiers:
                self.logger(event=f"Loading modifier {entry_point.name}")
                ret[entry_point.name] = entry_point.load()
        loaded = len(ret)
        configured = len(configured_modifiers)
        assert loaded == configured, (
            f"Number of modifiers loaded ({loaded})"
            f"differ from configured: {configured_modifiers}"
        )
        return ret

    def load_global_modifier_entrypoints(
        self, entry_points: Iterable[EntryPoint], configured_modifiers: List[str]
    ) -> Dict[str, Type[GlobalModifier]]:
        ret = dict()
        for entry_point in entry_points:
            if entry_point.name in configured_modifiers:
                self.logger(event=f"Loading global modifier {entry_point.name}")
                ret[entry_point.name] = entry_point.load()

        loaded = len(ret)
        configured = len(configured_modifiers)
        assert loaded == configured, (
            f"Number of global modifiers loaded ({loaded})"
            f"differ from configured: {configured_modifiers}"
        )
        return ret

    def apply_modifications(self, data: Optional[SourceData]) -> SourceData:
        with self.stats.timed("modifiers.apply_ms"):
            if data is None:
                data = self.source_data

            source_data = deepcopy(data)
            for scope, instances in source_data.scopes.items():
                for g in self.global_modifiers.values():
                    global_modifier = g(instances)
                    global_modifier.apply()
                    source_data.scopes[scope] = global_modifier.join()

                for instance in source_data.scopes[scope]:
                    for m in self.modifiers.values():
                        modifier = m(instance)
                        if modifier.match():
                            # Modifies the instance in-place
                            modifier.apply()
            return source_data

    def refresh(self) -> SourceData:
        self.stats.increment("sources.attempt")
        try:
            new = SourceData()
            for source in self.sources:
                new.scopes[source.scope].extend(source.get())
        except Exception as e:
            self.logger(
                event="Error while refreshing sources",
                traceback=[line for line in traceback.format_exc().split("\n")],
                error=e.__class__.__name__,
                detail=getattr(e, "detail", "-"),
            )
            self.stats.increment("sources.error")
            raise

        # Is the new data the same as what we currently have
        if new == getattr(self, "source_data", None):
            self.stats.increment("sources.unchanged")
            self.last_updated = datetime.now()
            return self.source_data
        else:
            self.stats.increment("sources.refreshed")
            self.last_updated = datetime.now()
            self.instance_count = len(
                [instance for scope in new.scopes.values() for instance in scope]
            )
            return new

    def extract_node_key(self, node: Union[Node, Dict[Any, Any]]) -> Any:
        if "." not in self.node_match_key:
            # key is not nested, don't need glom
            node_value = getattr(node, self.node_match_key)
        else:
            try:
                node_value = glom(node, self.node_match_key)
            except PathAccessError:
                raise RuntimeError(
                    f'Failed to find key "{self.node_match_key}" in discoveryRequest({node}).\n'
                    f"See the docs for more info: "
                    f"https://vsyrakis.bitbucket.io/sovereign/docs/html/guides/node_matching.html"
                )
        return node_value

    def extract_source_key(self, source: Dict[Any, Any]) -> Any:
        if "." not in self.source_match_key:
            # key is not nested, don't need glom
            source_value = source[self.source_match_key]
        else:
            try:
                source_value = glom(source, self.source_match_key)
            except PathAccessError:
                raise RuntimeError(
                    f'Failed to find key "{self.source_match_key}" in instance({source}).\n'
                    f"See the docs for more info: "
                    f"https://vsyrakis.bitbucket.io/sovereign/docs/html/guides/node_matching.html"
                )
        return source_value

    def match_node(
        self,
        node_value: Any,
        modify: bool = True,
    ) -> SourceData:
        """
        Checks a node against all sources, using the node_match_key and source_match_key
        to determine if the node should receive the source in its configuration.
        """

        if self.data_is_stale:
            # Log/emit metric and manually refresh sources.
            self.stats.increment("sources.stale")
            self.logger(
                event="Sources have not been refreshed in 2 minutes",
                last_update=self.last_updated,
                instance_count=self.instance_count,
            )
            self.poll()

        ret = SourceData()

        if modify:
            if not hasattr(self, "source_data_modified"):
                self.poll()
            data = self.source_data_modified
        else:
            data = self.source_data

        for scope, instances in data.scopes.items():
            if self.matching_enabled is False:
                ret.scopes[scope] = instances
                continue

            for instance in instances:
                source_value = self.extract_source_key(instance)

                # If a single expression evaluates true, the remaining are not evaluated/executed.
                # This saves (a small amount of) computation, which helps when the server starts
                # to receive thousands of requests. The list has been ordered descending by what
                # we think will more commonly be true.
                match = (
                    contains(source_value, node_value)
                    or node_value == source_value
                    or is_wildcard(node_value)
                    or is_wildcard(source_value)
                    or is_debug_request(node_value)
                )
                if match:
                    ret.scopes[scope].append(instance)
        return ret

    @property
    def match_keys(self) -> List[str]:
        """
        Checks for all match keys present in existing sources and adds them to a list

        A dict is used instead of a set because dicts cannot have duplicate keys, and
        have ordering since python 3.6
        """
        ret: Dict[str, None] = dict()
        ret["*"] = None
        for _, instances in self.source_data.scopes.items():
            if self.matching_enabled is False:

                break
            for instance in instances:
                source_value = glom(instance, self.source_match_key)
                if isinstance(source_value, str):
                    ret[source_value] = None
                elif isinstance(source_value, Iterable):
                    for item in source_value:
                        ret[item] = None
                    continue
                ret[source_value] = None
        return list(ret.keys())

    def poll(self) -> None:
        self.source_data = self.refresh()
        self.source_data_modified = self.apply_modifications(self.source_data)

    async def poll_forever(self) -> None:
        while True:
            self.poll()
            await asyncio.sleep(self.source_refresh_rate)



from typing import Dict

from fastapi import Body, Header
from fastapi.routing import APIRouter
from fastapi.responses import Response

from sovereign import discovery, logs
from sovereign.utils.auth import authenticate
from sovereign.schemas import (
    DiscoveryRequest,
    DiscoveryResponse,
    ProcessedTemplate,
)


router = APIRouter()

type_urls = {
    "v2": {
        "listeners": "type.googleapis.com/envoy.api.v2.Listener",
        "clusters": "type.googleapis.com/envoy.api.v2.Cluster",
        "endpoints": "type.googleapis.com/envoy.api.v2.ClusterLoadAssignment",
        "secrets": "type.googleapis.com/envoy.api.v2.auth.Secret",
        "routes": "type.googleapis.com/envoy.api.v2.RouteConfiguration",
        "scoped-routes": "type.googleapis.com/envoy.api.v2.ScopedRouteConfiguration",
    },
    "v3": {
        "listeners": "type.googleapis.com/envoy.config.listener.v3.Listener",
        "clusters": "type.googleapis.com/envoy.config.cluster.v3.Cluster",
        "endpoints": "type.googleapis.com/envoy.config.endpoint.v3.ClusterLoadAssignment",
        "secrets": "type.googleapis.com/envoy.extensions.transport_sockets.tls.v3.Secret",
        "routes": "type.googleapis.com/envoy.config.route.v3.RouteConfiguration",
        "scoped-routes": "type.googleapis.com/envoy.config.route.v3.ScopedRouteConfiguration",
        "runtime": "type.googleapis.com/envoy.service.runtime.v3.Runtime",
    },
}


def response_headers(
    discovery_request: DiscoveryRequest, response: ProcessedTemplate, xds: str
) -> Dict[str, str]:
    return {
        "X-Sovereign-Client-Build": discovery_request.envoy_version,
        "X-Sovereign-Client-Version": discovery_request.version_info,
        "X-Sovereign-Requested-Resources": ",".join(discovery_request.resource_names)
        or "all",
        "X-Sovereign-Requested-Type": xds,
        "X-Sovereign-Response-Version": response.version,
    }


@router.post(
    "/{version}/discovery:{xds_type}",
    summary="Envoy Discovery Service Endpoint",
    response_model=DiscoveryResponse,
    responses={
        200: {"description": "New resources provided"},
        304: {"description": "Resources are up-to-date"},
        404: {"description": "No resources found"},
    },
)
async def discovery_response(
    version: str,
    xds_type: str,
    discovery_request: DiscoveryRequest = Body(...),
    host: str = Header("no_host_provided"),
) -> Response:
    discovery_request.desired_controlplane = host
    response = await perform_discovery(
        discovery_request, version, xds_type, skip_auth=False
    )
    logs.queue_log_fields(
        XDS_RESOURCES=discovery_request.resource_names,
        XDS_ENVOY_VERSION=discovery_request.envoy_version,
        XDS_CLIENT_VERSION=discovery_request.version_info,
        XDS_SERVER_VERSION=response.version,
    )
    headers = response_headers(discovery_request, response, xds_type)

    if response.version == discovery_request.version_info:
        return not_modified(headers)
    elif getattr(response, "resources", None) == []:
        return Response(status_code=404, headers=headers)
    elif response.version != discovery_request.version_info:
        return Response(
            response.rendered, headers=headers, media_type="application/json"
        )
    return Response(content="Resources could not be determined", status_code=500)


async def perform_discovery(
    req: DiscoveryRequest,
    api_version: str,
    xds: str,
    skip_auth: bool = False,
) -> ProcessedTemplate:
    if not skip_auth:
        authenticate(req)
    try:
        type_url = type_urls[api_version][xds]
        req.type_url = type_url
    except (TypeError, KeyError):
        pass
    response = await discovery.response(req, xds)
    return response


def not_modified(headers: Dict[str, str]) -> Response:
    return Response(status_code=304, headers=headers)



from functools import partial
from collections import namedtuple
from typing import Optional, Any
from cryptography.fernet import Fernet, InvalidToken
from fastapi.exceptions import HTTPException


CipherSuite = namedtuple("CipherSuite", "encrypt decrypt key_available")


class DisabledSuite:
    @staticmethod
    def encrypt(_: bytes) -> bytes:
        return b"Unavailable (No Secret Key)"

    @staticmethod
    def decrypt(*_: bytes) -> str:
        return "Unavailable (No Secret Key)"


def create_cipher_suite(key: bytes, logger: Any) -> CipherSuite:
    try:
        fernet = Fernet(key)
        return CipherSuite(partial(encrypt, fernet), partial(decrypt, fernet), True)
    except TypeError:
        pass
    except ValueError as e:
        if key not in (b"", ""):
            logger.application_log(
                event=f"Fernet key was provided, but appears to be invalid: {repr(e)}"
            )
    return CipherSuite(DisabledSuite.encrypt, DisabledSuite.decrypt, False)


def generate_key() -> str:
    secret: bytes = Fernet.generate_key()
    return secret.decode()


def encrypt(cipher_suite: Fernet, data: str, key: Optional[str] = None) -> str:
    _local_cipher_suite = cipher_suite
    if isinstance(key, str):
        _local_cipher_suite = Fernet(key.encode())
    try:
        encrypted: bytes = _local_cipher_suite.encrypt(data.encode())
    except (InvalidToken, AttributeError):
        # TODO: defer this http error to later, return a normal error here
        raise HTTPException(status_code=400, detail="Encryption failed")
    return encrypted.decode()


def decrypt(cipher_suite: Fernet, data: str, key: Optional[str] = None) -> str:
    _local_cipher_suite = cipher_suite
    if key is not None:
        _local_cipher_suite = Fernet(key.encode())
    try:
        decrypted = _local_cipher_suite.decrypt(data.encode())
    except (InvalidToken, AttributeError):
        # TODO: defer this http error to later, return a normal error here
        raise HTTPException(status_code=400, detail="Decryption failed")
    if isinstance(decrypted, bytes):
        return decrypted.decode()
    else:
        return decrypted



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['sovereign',
 'sovereign.modifiers',
 'sovereign.sources',
 'sovereign.utils',
 'sovereign.views']

package_data = \
{'': ['*'], 'sovereign': ['static/*', 'static/sass/*', 'templates/*']}

install_requires = \
['Jinja2>=2.11.3,<3.0.0',
 'PyYAML>=5.4.1,<6.0.0',
 'aiofiles>=0.6.0,<0.7.0',
 'cachelib>=0.1.1,<0.2.0',
 'croniter>=1.3.5,<2.0.0',
 'cryptography>=3.3.1,<4.0.0',
 'fastapi>=0.63.0,<0.64.0',
 'glom>=20.11.0,<21.0.0',
 'gunicorn>=20.0.4,<21.0.0',
 'httptools>=0.1.1,<0.2.0',
 'requests>=2.25.1,<3.0.0',
 'schedule>=1.0.0,<2.0.0',
 'starlette==0.13.6',
 'structlog==20.1.0',
 'uvicorn>=0.13.3,<0.14.0',
 'uvloop>=0.16.0,<0.17.0']

extras_require = \
{'boto': ['boto3>=1.17.0,<2.0.0'],
 'orjson': ['orjson>=3.4.7,<4.0.0'],
 'sentry': ['sentry-sdk>=0.19.5,<0.20.0'],
 'statsd': ['datadog>=0.39.0,<0.40.0'],
 'ujson': ['ujson>=4.0.2,<5.0.0']}

entry_points = \
{'console_scripts': ['sovereign = sovereign.server:main'],
 'sovereign.modifiers': ['sovereign_3rd_party_test = '
                         'sovereign.modifiers.test:Test'],
 'sovereign.sources': ['file = sovereign.sources.file:File',
                       'inline = sovereign.sources.inline:Inline']}

setup_kwargs = {
    'name': 'sovereign',
    'version': '0.12.1',
    'description': 'Envoy Proxy control-plane written in Python',
    'long_description': 'sovereign\n=========\n\nMission statement\n-----------------\nThis project implements a JSON control-plane based on the [envoy](https://envoyproxy.io) [data-plane-api](https://github.com/envoyproxy/data-plane-api)\n\nThe purpose of `sovereign` is to supply downstream envoy proxies with \nconfiguration in near-realtime by responding to discovery requests.\n\nMechanism of Operation\n----------------------\ntl;dr version:\n```\n* Polls HTTP/File/Other for data\n* (optional) Applies transforms to the data\n* Uses the data to generate Envoy configuration from templates\n```\n\nIn a nutshell, Sovereign \ngathers contextual data (*"sources"* and *"template context"*), \noptionally applies transforms to that data (using *"modifiers"*) and finally \nuses the data to generate envoy configuration from either python code, or jinja2 templates.\n\nThis is performed in a semi-stateless way, where the only state is data cached in memory.\n\nTemplate context is intended to be statically configured, whereas *Sources* \nare meant to be dynamic - for example, fetching from an API, an S3 bucket, \nor a file that receives updates.\n\n*Modifiers* can mutate the data retrieved from sources, just in case the data \nis in a less than favorable structure.\n\nBoth modifiers and sources are pluggable, i.e. it\'s easy to write your own and \nplug them into Sovereign for your use-case.\n\nCurrently, Sovereign supports only providing configuration to Envoy as JSON. \nThat is to say, gRPC is not supported yet. Contributions in this area are highly\nappreciated!\n\nThe JSON configuration can be viewed in real-time with Sovereign\'s read-only web interface.\n\nRequirements\n------------\n* Python 3.8+\n\nInstallation\n------------\n```\npip install sovereign\n```\n\nDocumentation\n-------------\n[Read the docs here!](https://vsyrakis.bitbucket.io/sovereign/docs/)\n\n:new: Read-only user interface\n------------------------\nAdded in `v0.5.3`!\n\nThis interface allows you to browse the resources currently returned by Sovereign.\n\n![Sovereign User Interface Screenshot](https://bitbucket.org/atlassian/sovereign/src/master/assets/sovereign_ui.png)\n\nLocal development\n=================\n\nRequirements\n------------\n* Docker\n* Docker-compose\n\nInstalling dependencies for dev\n-------------------------------\nI recommend creating a virtualenv before doing any dev work\n\n```\npython3 -m venv venv\nsource venv/bin/activate\npip install -r requirements-dev.txt\n```\n\nRunning locally\n---------------\nRunning the test env\n\n```\nmake run\n```\n    \nRunning the test env daemonized\n\n```\nmake run-daemon\n```\n\nPylint\n\n```\nmake lint\n```\n\nUnit tests\n\n```\nmake unit\n```\n\nAcceptance tests\n\n```\nmake run-daemon acceptance\n```\n\n\nContributors\n============\n\nPull requests, issues and comments welcome. For pull requests:\n\n* Add tests for new features and bug fixes\n* Follow the existing style\n* Separate unrelated changes into multiple pull requests\n\nSee the existing issues for things to start contributing.\n\nFor bigger changes, make sure you start a discussion first by creating\nan issue and explaining the intended change.\n\nAtlassian requires contributors to sign a Contributor License Agreement,\nknown as a CLA. This serves as a record stating that the contributor is\nentitled to contribute the code/documentation/translation to the project\nand is willing to have it used in distributions and derivative works\n(or is willing to transfer ownership).\n\nPrior to accepting your contributions we ask that you please follow the appropriate\nlink below to digitally sign the CLA. The Corporate CLA is for those who are\ncontributing as a member of an organization and the individual CLA is for\nthose contributing as an individual.\n\n* [CLA for corporate contributors](https://na2.docusign.net/Member/PowerFormSigning.aspx?PowerFormId=e1c17c66-ca4d-4aab-a953-2c231af4a20b)\n* [CLA for individuals](https://na2.docusign.net/Member/PowerFormSigning.aspx?PowerFormId=3f94fbdc-2fbe-46ac-b14c-5d152700ae5d)\n\n\nLicense\n========\n\nCopyright (c) 2018 Atlassian and others.\nApache 2.0 licensed, see [LICENSE.txt](LICENSE.txt) file.\n\n\n',
    'author': 'Vasili Syrakis',
    'author_email': 'vsyrakis@atlassian.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pypi.org/project/sovereign/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import gunicorn.app.base
from fastapi import FastAPI
from typing import Optional, Dict, Any
from sovereign import asgi_config
from sovereign.app import app


class StandaloneApplication(gunicorn.app.base.BaseApplication):  # type: ignore
    def __init__(
        self, application: FastAPI, options: Optional[Dict[str, Any]] = None
    ) -> None:
        self.options = options or {}
        self.application = application
        super().__init__()

    def load_config(self) -> None:
        for key, value in self.options.items():
            self.cfg.set(key.lower(), value)

    def load(self) -> FastAPI:
        return self.application


def main() -> None:
    asgi = StandaloneApplication(
        application=app, options=asgi_config.as_gunicorn_conf()
    )
    asgi.run()


if __name__ == "__main__":
    main()
